import org.gjt.sp.jedit.*;
import org.gjt.sp.jedit.textarea.*;
import javax.swing.*;
import java.util.*;
import javax.swing.text.Element;
import javax.swing.text.Position;
import java.awt.event.ActionEvent;
import javax.swing.text.BadLocationException;

public class complete_abbreviation extends EditAction {
	
	public complete_abbreviation() {
		
		super("complete-abbreviation");
	}
	
	private JEditTextArea textArea;
	private Mode mode;
	private View view;
	private Buffer buffer;
	private Position position;
	private int lineNo;
	private int selectionStart;
	
	public void actionPerformed(ActionEvent evt) {
	
		view = getView(evt);
		buffer = view.getBuffer();
		textArea = view.getTextArea();
		mode = buffer.getMode();

		String separators = (String)buffer.getProperty("noWordSep");
		String line;
				
		int dot = textArea.getSelectionStart();
		lineNo = textArea.getSelectionStartLine();	
		int start = textArea.getLineStartOffset(lineNo);
		int len = textArea.getLineEndOffset(lineNo) - start - 1;
		
		if (dot == start) {
			return;
		}
		
		try {
			line = buffer.getText(start,len); // chop newline
			char currentChar = line.charAt(dot - start - 1);
			String charAbbreviated = new Character(currentChar).toString();
			String charExpanded = expandAbbrev(charAbbreviated, false);
			if (charExpanded != null) {
				selectionStart = dot - 1;
				textArea.setSelectionStart(selectionStart);
				insertString(charExpanded);
				return;
			}
			
			// scan backwards to find word
			int wordStart = start;
loop:		for(int i = dot - 1; i >= start; i--) {
				char c = line.charAt(i - start);
				switch(c) {
					case ' ':
					case '\t':
					wordStart = i + 1;
					break loop;
					default:
					if(!Character.isLetterOrDigit(c)
						&& (separators == null
						|| separators.indexOf(c) == -1)) {
						wordStart = i + 1;
						break loop;
					}
				}
			}
			if(wordStart == dot) {
				view.getToolkit().beep();
				System.out.println("wordStart == dot beep");
				return;
			}
			String stringAbbreviated = line.substring(wordStart - start,
						     dot - start);
			String stringExpanded = expandAbbrev(stringAbbreviated, true);
			if (stringExpanded == null) {
				view.getToolkit().beep();
				System.out.println("stringExpanded == null beep");
				return;
			}
			selectionStart = dot - stringAbbreviated.length();
			textArea.setSelectionStart(selectionStart);
			insertString(stringExpanded);
		} catch(BadLocationException bl) {}
	}
	
	private String expandAbbrev(String abbrev, boolean add) {
//		System.out.println(abbrev);
		String expand = Abbrevs.expandAbbrev(buffer.getMode().getName(), abbrev);
		if(add == true && expand == null) {
			return TSAbbrevs.addAbbrev(view, abbrev);
		}
		return expand;
//		return PropertiesLoader.myProperties.getProperty(abbrev);
	}
	
	private void insertString(String data) {

		/* Number of initial tabs and spaces per line. numberOfTabs[0] means 
		the first line etc. */
		int[] numberOfTabs;
		int[] numberOfSpaces;

		/* Convert "\\n" to "\n" and "\\t" to "\t" to make using 
		StringTokenizer possible. */
		data = convertEscapeCodes(data);

		/* Delimiter is '\n' only. Newlines are separate tokens. */
		StringTokenizer lines = new StringTokenizer(data, "\n", true);

		/* lines.countTokens() is usually not the actual number of lines, but
		it is the maximum. */
		numberOfTabs = new int[lines.countTokens()];
		numberOfSpaces = new int[lines.countTokens()];

		/* data will be reconstructed. */
		data = "";

		int currentLine = 0;

		while (lines.hasMoreTokens()) {

			String line = new String(lines.nextToken());

			/* Append the delimiter, but only if we're not handling an empty 
			line. */
			if (!line.equals("\n") && lines.hasMoreTokens()) {
				line += lines.nextToken();
			}

			char c = ' ';

			while ( (line.length()>0) && (c=='\t' || c==' ') ) {

				c=line.charAt(0);

				switch (c) {
					case ' ': {
						numberOfSpaces[currentLine]++;
						line = line.substring(1);
						break;
					}
					case '\t': {
						numberOfTabs[currentLine]++;
						line = line.substring(1);
						break;
					}
				}

			}

			data = data + line;
			currentLine++;
		}
		

		int j = 0;
		int pos = -1;
		int ln = 0;
		int caretLine = 0;
		char c = '\0';
		int dataLength = data.length();
		StringBuffer buf = new StringBuffer(dataLength);
		
		for (int i = 0; i < dataLength; i++) {
			switch(c = data.charAt(i)) {
				case '|':
					if (i < (dataLength - 1) && data.charAt(i + 1) == '|') {
						i++;
						buf.append(c);
					} else {
						if (pos != -1) {
							break;
						}
						pos = j;
						caretLine = ln;
//						System.out.println("caretLine = " + caretLine);
					}
					break;
				case '\n':
					ln++;
//					System.out.println("ln++ from second " + ln);
					buf.append(c);
					break;
				default:
					buf.append(c);
					break;
			}
			j++;
		}

		/* If caret position isn't set, set it to the end of pasted text. */
		if (pos == -1) {
			pos = j;
			caretLine = ln;
		}

		textArea.setSelectedText(buf.toString());
		int endLineNo = textArea.getSelectionEndLine();
		try {
			position = buffer.createPosition(selectionStart + pos);
		} catch(BadLocationException bl) {}

		if (ln > 0) {
			int between = endLineNo - lineNo;
//			System.out.println("between = " + endLineNo + "-" + lineNo);
			for (int i = between; i >= 0; i--) {
				buffer.indentLine(textArea, endLineNo - i, false);				
//				System.out.println((endLineNo - i) + " line indented"); 
			}

			/* Insert leading whitespace. It is not inserted into the first 
			line. */
			for (int i = lineNo+1; i <= endLineNo; i++) {
					insertWhitespace(i, numberOfTabs[i-lineNo],
									numberOfSpaces[i-lineNo]);
			}

		}

		/* Finally set the caret position. */
		textArea.setCaretPosition(position.getOffset());

	}

//	if we are using Marker or Register we can't replace the entire line: this
//	causes the caret to be placed at the end of the line.
//	When using Position we can replace the entire line
	/**
	 * Inserts tabs and spaces into the beginning of given line. Spaces are
	 * inserted after tabs.
	 * @param line		The line number
	 * @param tabs		Number of tabs to insert
	 * @param spaces	Number of spaces to insert
	 * @author Kristian Ovaska, Valery Kondakoff
	*/
	private void insertWhitespace(int line, int tabs, int spaces) {

		if ( (tabs==0) && (spaces==0) ) {
			return;
		}

		int lineStartOffset = textArea.getLineStartOffset(line);
		
		textArea.select(lineStartOffset, lineStartOffset + 
									textArea.getLineLength(line));

		String lineText = textArea.getSelectedText();
		if (lineText==null) {
			lineText = "";
		}

		StringBuffer lineBuffer = new StringBuffer(lineText);
//		System.out.println(useMarker + " " + lineBuffer + " line = " + line);

		/* Insert tabs. */
		for (int i=0; i<tabs; i++) {
			lineBuffer.insert(0, '\t');			
		}

		/* Insert spaces. They go after the tabs this line possibly contains. */
		if (spaces>0) {

			int pos = 0;

			while ( (pos<lineBuffer.length()) && (lineBuffer.charAt(pos)=='\t') ) {
				pos++;
			}

			for (int i=0; i<spaces; i++) {
				lineBuffer.insert(pos, ' ');
			}

		}

		textArea.setSelectedText(lineBuffer.toString());

	}

	
//	we need to convert "\\n" _and_ "\\t"
	/**
	 * Converts all "\\n" to "\n".
	 * Converts all "\\t" to "\t"
	 * @author Kristian Ovaska, Valery Kondakoff
	*/
	private String convertEscapeCodes(String s) {

		StringBuffer converted = new StringBuffer(s.length());

		for (int i=0; i<s.length(); i++) {
			
			if ( (i<s.length()-1) && (s.charAt(i)=='\\') ) {
			
				char c = s.charAt(i + 1);
				
				switch (c) {
					case 'n':
						converted.append('\n');
						i++; /* Skip 'n'. */
//						System.out.println("case n");
						break;
					case 't':
						converted.append('\t');
						i++; /* Skip 't'. */
//						System.out.println("case t");
						break;
				}
				
			} else {
				converted.append(s.charAt(i));
			}
		}

		return converted.toString();
	}

}



