//behavior: 
//	- this class comments/uncomments full or partially selected lines;
//	- this class comments/uncomments entire lines, not parts of lines. 
//	  Therefore comments in the middle of the line are not
//	  added/removed. Instead entire line will be commented/uncommented;
//notes: 
//	- comment is inserted at the beginning of the line (for better
//     code organisation);	
//	- you don't need to select entire line to comment/uncomment it. 
//	   Select some part of the line or just set Caret somewhere within
//	   the line;

import javax.swing.text.BadLocationException;
import javax.swing.text.Element;
import java.awt.event.ActionEvent;
import org.gjt.sp.jedit.textarea.JEditTextArea;
import org.gjt.sp.util.Log;
import org.gjt.sp.jedit.*;
import gnu.regexp.*;

public class line_comment extends EditAction {
	public line_comment() {
		super("line-comment");
	}
	public void actionPerformed(ActionEvent evt) {
		View view = getView(evt);
		JEditTextArea textArea = view.getTextArea();
		Buffer buffer = view.getBuffer();
		String comment = (String)buffer.getProperty("blockComment");
		String regExpString = null;
		if (comment == null) {
			view.getToolkit().beep();
			return;
		}
		if (comment.equals("//") || comment.equals("#") ||
				comment.equals("--") || comment.equals("%") ||
				comment.equals("rem")) {
			regExpString = "^\\s*(" + comment + ")";
		} else {
			char[] commentArray = comment.toCharArray();
			StringBuffer sbComment = new StringBuffer(comment);
			for (int i = 0; i < commentArray.length; i++) {
				switch (commentArray[i]) {
					case '.':  //through
					case '?':  //through
					case '*':  //through
					case '(':  //through
					case ')':  //through
					case '|':  //through
					case '[':  //through
					case ']':  //through
					case '\\':
						sbComment.insert(i, '\\');
						commentArray = (sbComment.toString()).toCharArray();
						i++;
						break;
					default:
						break;
				}
			}
			regExpString = "^\\s*(" + sbComment.toString() + ")";
		}
		RE regExp = null;
		int startLine = textArea.getSelectionStartLine();
		int endLine = textArea.getSelectionEndLine();
		int commentLength = comment.length();
		try {
			regExp = new RE(regExpString);
		} catch (REException e) {
			Log.log(Log.ERROR, this, e);
			return;
		}
		Element map = buffer.getDefaultRootElement();
		buffer.beginCompoundEdit();
		try {
			for(int i = startLine; i <= endLine; i++) {
				int offset = map.getElement(i).getStartOffset();
				REMatch match = regExp.getMatch(textArea.getLineText(i));
				if (match != null) {
					int matchEnd = match.getEndIndex();
					buffer.remove(offset, matchEnd);
					buffer.insertString(offset,match.toString().substring(0, 
											(matchEnd - commentLength)),null);
					continue;
				}
				buffer.insertString(offset,comment,null);
			}
		}
		catch(BadLocationException bl) {
			Log.log(Log.ERROR,this,bl);
		}
		finally {
			buffer.endCompoundEdit();
		}
		textArea.select(textArea.getCaretPosition(),
			textArea.getCaretPosition());
	}
}
// last updated nov 25, 1999, 0:49
